use std::sync::Arc;

use rustls::{
    crypto::aws_lc_rs::sign::any_supported_type, 
    pki_types::{pem::PemObject, CertificateDer, PrivateKeyDer}, 
    server::{ClientHello, ResolvesServerCert}, 
    sign::CertifiedKey, 
    ServerConfig
};

use super::config::Config;


#[derive(Clone, Debug)]
pub struct TlsCertificate {
    key: CertifiedKey
}

impl TlsCertificate {
    pub fn new(cert_file: &str, key_file: &str) -> Option<TlsCertificate> {
        let certs = CertificateDer::pem_file_iter(cert_file)
            .unwrap()
            .map(|cert| cert.unwrap())
            .collect();
        let private_key = PrivateKeyDer::from_pem_file(key_file).unwrap();
        let key = CertifiedKey::new(certs, any_supported_type(&private_key).ok()?);

        Some(Self { key })
    }

    pub fn get_key(&self) -> CertifiedKey {
        self.key.clone()
    }
}

#[derive(Debug)]
struct ResolvesServerCertWildcard {
    config: Arc<Config>
}

impl ResolvesServerCertWildcard {
    pub fn new(config: Arc<Config>) -> Self {
        Self { config }
    }
}

impl ResolvesServerCert for ResolvesServerCertWildcard {
    fn resolve(&self, client_hello: ClientHello<'_>) -> Option<Arc<CertifiedKey>> {
        if let Some(cert) = client_hello.server_name()
                .and_then(|name| self.config.get_site(name).cloned())
                .and_then(|site| site.ssl) {
            Some(Arc::new(cert.get_key()))
        } else {
            None
        }
    }
}

pub fn create_server_config(config: Arc<Config>) -> ServerConfig {
    ServerConfig::builder()
        .with_no_client_auth()
        .with_cert_resolver(Arc::new(ResolvesServerCertWildcard::new(config)))
}